import six
import inspect


def get_not_implemented(cls_name, cls, method):
    return NotImplementedError('{0} (a subclass of {1}) must override method {1}.{2}'.format(
        cls_name,
        cls.__name__,
        method
    ))


class InterfaceMeta(type):
    def __new__(cls, name, parents, dct):
        parent_interface = next((i for i in parents if hasattr(i, '__interface__')), None)

        if not parent_interface:
            dct['__interface__'] = True
        else:
            for key, value in inspect.getmembers(parent_interface, inspect.isfunction):
                if key not in dct:
                    raise get_not_implemented(name, parent_interface, key)

        return super(InterfaceMeta, cls).__new__(cls, name, parents, dct)


class Interface(six.with_metaclass(InterfaceMeta)):
    pass



import json, uuid, logging, thread, time, os

log = logging.getLogger(__name__)

_TASKS = {}
_manager = None


def task(*args, **kwargs):
    """
    Decorator which creates and registers a new task
    """
    def deco(f):
        task = Task(f.__name__, f, *args, **kwargs)

        if f.__name__ in _TASKS:
            raise Exception("Conflicting task name: %s" % f.__name__)

        _TASKS[f.__name__] = task
        return task
    return deco


class Task(object):
    def __init__(self, name, f, max_running=None, buffer_time=None, max_size=25):
        self.name = name
        self.f = f
        self.max_running = max_running
        self.max_size = max_size
        self.buffer_time = buffer_time

    def __call__(self, *args, **kwargs):
        return self.f(*args, **kwargs)

    def queue(self, *args, **kwargs):
        if _manager.get_queue_length(self.name) > self.max_size:
            raise Exception('Queue for task {} is full'.format(self.name))

        return _manager.add_job(self.name, {
            'id': id,
            'args': args,
            'kwargs': kwargs,
        })


class TaskRunner(object):
    def __init__(self, manager, name, task):
        self.manager = manager
        self.name = name
        self.f = task
        self.running = 0

    def process(self, job):
        self.running += 1
        log.info('[%s] Running job %s...', job['id'], self.name)
        self.manager.track_job_start(job)
        start = time.time()

        try:
            self.f(*job['args'], **job['kwargs'])
            if self.f.buffer_time:
                time.sleep(self.f.buffer_time)
        except:
            log.exception("[%s] Failed in %ss", job['id'], time.time() - start)
        finally:
            self.manager.track_job_end(job)
            self.running -= 1

        log.info('[%s] Completed in %ss', job['id'], time.time() - start)

    def run(self, job):
        if self.f.max_running:
            while self.f.max_running <= self.running:
                time.sleep(.5)

        thread.start_new_thread(self.process, (job, ))


class TaskManager(object):
    def __init__(self, redis=None, autoload=True):
        # If no global manager has been created, set us to it
        global _manager
        if not _manager:
            _manager = self

        self.redis = redis
        self.queue = None
        self.runners = {k: TaskRunner(self, k, v) for k, v in _TASKS.items()}
        self.active = True

        # Load tasks modules to register tasks
        if autoload:
            self.load()

    def add_job(self, name, job):
        id = str(uuid.uuid4())

        if self.redis:
            self.redis.rpush('jq:%s' % name, json.dumps(job))
        else:
            self.queue.put((name, job))

        return id

    def track_job_start(self, job):
        self.redis.set("task:%s" % job['id'], 1)

    def track_job_end(self, job):
        self.redis.delete("task:%s" % job['id'])

    def load(self):
        for f in os.listdir("tasks/"):
            if f.endswith(".py"):
                __import__("tasks." + f.rsplit(".")[0])

    def run_forever(self):
        from gevent import spawn
        spawn(self.run)

    def run_redis(self):
        queues = ['jq:' + i for i in _TASKS.keys()]

        while self.active:
            chan, job = self.redis.blpop(queues)
            job_name = chan.split(":", 1)[1]
            job = json.loads(job)

            if job_name not in _TASKS:
                log.error('Cannot handle task %s', job_name)
                continue

            self.runners[job_name].run(job)

    def run_gevent(self):
        import gevent
        self.queue = gevent.queue.Queue()

        while self.active:
            job_name, job = self.queue.get()

            if job_name not in self.runners:
                log.error('Cannot handle task %s', job_name)
                continue

            gevent.spawn(self.runners[job_name].run, job)

    def run(self):
        log.info("Running TaskManager on %s queues...", len(self.queues))

        if self.redis:
            return self.run_redis()
        else:
            return self.run_gevent()



import sys, os

from flask import current_app

from .filters import register_filters
from .log import setup_logging
from .session import SessionProvider


class ViewFinder(object):
    @staticmethod
    def get_views_path():
        """
        Walks current path to find the views directory
        """
        for root, dirs, files in os.walk(os.getcwd()):
            if root.endswith("views"):
                return root

    @classmethod
    def get_views(cls):
        """
        Returns all views that should be loaded
        """
        views_path = cls.get_views_path()
        if not views_path:
            raise Exception('Failed to locate views directory')

        module = views_path.replace(os.getcwd(), '').replace('/', '.')[1:]
        for view in os.listdir(views_path):
            if not view.endswith(".py") or view.startswith("_"):
                continue

            view = module + "." + view.split(".py")[0]
            __import__(view)
            yield getattr(sys.modules[view], view.split(".")[-1])


class Holster(object):
    def __init__(self, app=None, auto_load_views=True, use_sessions=False):
        self.auto_load_views = auto_load_views

        self.app = app or current_app
        if self.app:
            self.init_app(self.app)

        # Optionally initialize redis
        self.redis = None

        # Initialize redis if its in our app config
        if 'REDIS' in self.app.config:
            self.init_redis(self.app.config.get("REDIS"))

        # Initialize sessions if we have redis and want to use them
        if self.redis and use_sessions:
            self.sessions = SessionProvider(self.redis)
            self.app.sessions = self.sessions

    def init_app(self, app):
        app.holster = self
        register_filters(app)
        setup_logging(app)

        if self.auto_load_views:
            map(self.app.register_blueprint, ViewFinder.get_views())

    def init_redis(self, kwargs):
        from redis import Redis
        self.redis = Redis(**kwargs)



from flask import flash, redirect


def flashy(msg, etype="danger", path="/"):
    """
    Flashes a message and redirects a user. Returns a flask Response that can
    be returned to the user.
    """
    flash(msg, etype)
    return redirect(path)


class SimpleObject(object):
    def __init__(self, data=None):
        self.__dict__.update(data or {})



from collections import OrderedDict

import six


class EnumAttr(object):
    def __init__(self, parent, name, index, value):
        self.parent = parent
        self.name = name
        self.index = index
        self.value = value

    def __eq__(self, other):
        if isinstance(other, EnumAttr):
            return (self.parent == other.parent) and (self.index == other.index)

        return self.value == other

    def __cmp__(self, other):
        if isinstance(other, EnumAttr):
            return self.index - other.index

        return self.value.__cmp__(other)

    def __repr__(self):
        return '<EnumAttr {}>'.format(self.name)

    def __str__(self):
        return self.name

    def __int__(self):
        return self.index

    def __hash__(self):
        return hash((self.name, self.index, self.value))


class BaseEnumMeta(type):
    def __getattr__(self, attr):
        if attr.lower() in self.attrs:
            return self.attrs[attr.lower()]
        raise AttributeError

    def __getitem__(self, item):
        return self.get(item)

    def get(self, entry):
        for attr in six.itervalues(self.attrs):
            if attr == entry or attr.name == entry or attr.value == entry:
                return attr

    @property
    def ALL(self):
        return set(self.attrs.keys())

    @property
    def ALL_VALUES(self):
        return set(i.value for i in six.itervalues(self.attrs))


def bitmask_enumerate(seq):
    for i, e in enumerate(seq):
        yield (1 << i), e


def Enum(*args, **kwargs):
    class _T(six.with_metaclass(BaseEnumMeta)):
        pass

    _T.attrs = OrderedDict()
    _T.order = []

    if args:
        enumer = enumerate
        if kwargs.get('bitmask', True):
            enumer = bitmask_enumerate

        _T.order = args
        _T.attrs = {e.lower(): EnumAttr(_T, e.lower(), i, e) for i, e in enumer(args)}
    else:
        _T.order = []
        _T.attrs = {k.lower(): EnumAttr(_T, k.lower(), v, v) for k, v in kwargs.items()}

    return _T



from __future__ import absolute_import

import logging, os

LEVELS = {
    "urllib3": logging.WARNING,
    "requests": logging.DEBUG,
}

FORMAT = "[%(levelname)s] %(asctime)s - %(name)s:%(lineno)d - %(message)s"


def set_logging_levels():
    for log, lvl in LEVELS.items():
        logging.getLogger(log).setLevel(lvl)


def setup_logging(app):
    logging.basicConfig(level=logging.DEBUG, format=FORMAT)
    set_logging_levels()

    var_log_path = "/var/log/{}".format(app.logger_name)

    if os.path.exists(var_log_path):
        file_handler = logging.FileHandler(os.path.join(var_log_path, "flask.log"))
    else:
        file_handler = logging.FileHandler('/tmp/{}-flask.log'.format(app.logger_name))

    file_handler.setFormatter(logging.Formatter(FORMAT))

    root = logging.getLogger()
    root.addHandler(file_handler)



import json


def to_json_filter(data):
    return json.dumps(data, separators=(',', ':'))


def to_pjson_filter(data):
    return json.dumps(data)


def register_filters(app):
    app.add_template_filter(to_json_filter, "to_json")
    app.add_template_filter(to_pjson_filter, "to_pjson")




import time, logging, signal, sys, os

from flask import current_app

from dateutil.relativedelta import relativedelta
from datetime import datetime

log = logging.getLogger(__name__)


class Task(object):
    def __init__(self, task, now, delta):
        self.redis = current_app.holster.redis
        self.task = task
        self.delta = delta
        self.last = (datetime.utcnow() - relativedelta(years=5)) if now else datetime.utcnow()
        self.last_id = None
        self.last_check = time.time()

    def is_running(self):
        if self.last_id:
            if (time.time() - self.last_check) < 5:
                return True

            self.last_check = time.time()
            res = bool(self.redis.exists("task:%s" % self.last_id))
            if not res:
                self.last_id = None
            return res

    def is_active(self):
        return True

    def should_run(self):
        return (self.is_active() and not self.is_running() and
            (datetime.utcnow() > (self.last + self.delta)))

    def run(self):
        log.info("Queueing scheduled task %s" % self.task.name)
        self.last_id = self.task.queue()
        self.last = datetime.utcnow()


class Scheduler(object):
    def __init__(self, pidfile="/tmp/scheduler.pid"):
        self.redis = current_app.holster.redis
        # Write PID to a file
        try:
            with open(pidfile, "w") as f:
                f.write(str(os.getpid()))
        except:
            log.exception("Failed to create PID file: ")

        self.schedules = []
        self.paused = False

        signal.signal(signal.SIGINT, self.handle_signal)
        signal.signal(signal.SIGTERM, self.handle_signal)
        signal.signal(signal.SIGHUP, self.handle_signal)

    def handle_signal(self, signum, handler):
        log.info("Handling signal %s" % signum)
        sys.exit(0)

    def schedule(self, f, start_now=False, **kwargs):
        self.schedules.append(Task(f, start_now, relativedelta(**kwargs)))

    def run(self):
        log.info("Starting Scheduler...")
        while True:
            time.sleep(.3)
            if self.paused:
                continue

            for task in self.schedules:
                if task.should_run():
                    task.run()



from collections import defaultdict

from .enum import Enum

Priority = Enum(
        'BEFORE',
        'NONE',
        'AFTER',
)


class Event(object):
    def __init__(self, parent, data):
        self.parent = parent
        self.data = data

    def __getattr__(self, name):
        if hasattr(self.data, name):
            return getattr(self.data, name)
        raise AttributeError


class EmitterSubscription(object):
    def __init__(self, events, prio, func):
        self.events = events
        self.prio = prio
        self.func = func
        self.emitter = None

    def add(self, emitter):
        self.emitter = emitter
        for event in self.events:
            emitter.event_handlers[self.prio][event].append(self)
        return self

    def remove(self, emitter=None):
        emitter = emitter or self.emitter

        for event in self.events:
            emitter.event_handlers[self.prio][event].remove(self)


class Emitter(object):
    def __init__(self, wrapper=None):
        self.wrapper = wrapper
        self.event_handlers = {
            k: defaultdict(list) for k in Priority.attrs.values()
        }

    def _call(self, func, args, kwargs):
        if self.wrapper:
            self.wrapper(func, *args, **kwargs)
        else:
            func(*args, **kwargs)

    def emit(self, name, *args, **kwargs):
        for prio in [Priority.BEFORE, Priority.NONE, Priority.AFTER]:
            for listener in self.event_handlers[prio].get(name, []):
                self._call(listener.func, args, kwargs)

    def on(self, *args, **kwargs):
        return EmitterSubscription(args[:-1], kwargs.pop('priority', Priority.NONE), args[-1]).add(self)



import json, time, logging

from flask import current_app

from uuid import uuid4
from flask import request

log = logging.getLogger(__name__)

# 5 days
DEFAULT_SESSION_TTL = 60 * 60 * 24 * 5


def find_sessions_for_user(uid):
    """
    Attempts to find all sessions associated with a user.
    """
    for key in current_app.holster.redis.keys("session:*"):
        sess = Session(key.split(":")[1])
        if sess['u'] == uid:
            yield sess


class SessionProvider(object):
    def __init__(self, redis, ttl=DEFAULT_SESSION_TTL):
        self.redis = redis
        self.ttl = ttl

    def get(self, id=None):
        return Session(self, id)


class Session(object):
    def __init__(self, provider, id=None):
        self.provider = provider
        self.prefix = "session:{}:{}".format(provider, '{}')
        self._id = None
        self._data = {}
        self._changed = False

        # Try to load a previous session
        id = id or request.cookies.get("s")
        data = self.provider.redis.get(self.prefix.format(id))
        if data:
            self._data = json.loads(data)
            self._id = id

    @property
    def key(self):
        return self.prefix.format(self._id)

    @property
    def new(self):
        return not self._id

    def delete(self):
        self.provider.redis.delete(self.key)
        self._data = {}
        self._id = None
        self._changed = False

    def get(self, key, default=None):
        if key in self._data:
            return self[key]
        return default

    def __delitem__(self, item):
        self._changed = True
        del self._data[item]

    def __getitem__(self, item):
        return self._data[item]

    def __setitem__(self, item, value):
        self._changed = True
        self._data[item] = value

    def save(self, response):
        if not self._changed:
            return False

        # Get current TTL
        self._id = self._id or str(uuid4())
        ttl = int(self.provider.redis.ttl(self.key) or self.provider.ttl)

        # Set cookie and key
        self.provider.redis.set(self.key, json.dumps(self._data))
        self.provider.redis.expire(self.key, ttl if ttl != -1 else self.provider.ttl)
        response.set_cookie("s", self._id, expires=(time.time() + ttl))
        return True



import json

from flask import Response, request


class APIResponse(Response):
    def __init__(self, obj={}):
        Response.__init__(self)

        if 'success' not in obj:
            obj['success'] = True

        if request.values.get("pretty") == '1':
            self.data = json.dumps(obj,
                sort_keys=True,
                indent=2,
                separators=(',', ': '))
        else:
            self.data = json.dumps(obj)

        self.mimetype = "application/json"



import itsdangerous

_unsafe_deserializer = itsdangerous.URLSafeTimedSerializer('')


class SecureEntity(object):
    def get_token_secret_key(self):
        raise NotImplementedError('{} must override SecureEntity.get_token_secret_key'.format(self.__class__.__name__))


def create_token(entity, contents=None, field=None, salt=b'itsdangerous'):
    assert isinstance(entity, SecureEntity)

    if callable(contents):
        contents = contents(entity)

    if not contents and field:
        contents = getattr(entity, field)

    return itsdangerous.URLSafeTimedSerializer(secret_key=entity.get_token_secret_key(), salt=salt).dumps(contents)


def peak_token_data(token):
    """
    Peaks at a tokens data without any validation/verification of the signature.
    :param token: The token to peak at.
    :rtype: Any valid JSON data
    """
    try:
        payload = itsdangerous.want_bytes(token).rsplit('.', 2)[0]
        return _unsafe_deserializer.load_payload(payload)
    except (itsdangerous.BadPayload, IndexError):
        return None


def verify_token(token, entity_getter, salt=b'itsdangerous', max_age=None):
    data = peak_token_data(token)
    if not data:
        return None, None

    entity = entity_getter(data)
    if not entity:
        return None, data

    try:
        itsdangerous.URLSafeTimedSerializer(entity.get_token_secret_key(), salt).loads(token, max_age=max_age)
    except (itsdangerous.BadSignature, itsdangerous.SignatureExpired):
        return None, None

    return entity, data



from flask import jsonify

from .util import flashy


class ResponseException(Exception):
    """
    Represents an exception that can be cleanly handled and displayed to the
    user. This exception does not imply an error has occured, but rather can
    be used as an easy way to break execution and return responses inside
    complicated view handlers.
    """
    def to_response(self):
        """
        Subclasses must implement this method. Returns a flask Response object.
        """
        raise NotImplementedError("Must define to_response on `%s`" % self.__class__.__name__)


class UserError(ResponseException):
    """
    Flashes a message and redirects a user.
    """
    def __init__(self, msg, etype="danger", path="/"):
        self.msg = msg
        self.etype = etype
        self.path = path

    def to_response(self):
        return flashy(self.msg, self.etype, self.path)


class APIError(ResponseException):
    def __init__(self, msg, status_code=200):
        self.msg = msg
        self.status_code = status_code

    def to_response(self):
        resp = jsonify({
            "message": self.msg,
            "success": False
        })
        resp.status_code = self.status_code
        return resp

    @classmethod
    def ensure(condition, message):
        if not condition:
            raise APIError(message)



import gevent
import weakref


class ThreadLocal(object):
    ___slots__ = ['storage']

    def __init__(self):
        self.storage = weakref.WeakKeyDictionary()

    def get(self):
        if gevent.getcurrent() not in self.storage:
            self.storage[gevent.getcurrent()] = {}
        return self.storage[gevent.getcurrent()]

    def drop(self):
        if gevent.getcurrent() in self.storage:
            del self.storage[gevent.getcurrent()]

    def __getitem__(self, item):
        return self.get()[item]

    def __setitem__(self, item, value):
        self.get()[item] = value



#!/usr/bin/env python

import os
import sys

import holster

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

packages = [
    'holster',
]

requires = [
    'six'
]

setup(
    name='holster',
    version=holster.__version__,
    description='a set of Python utilities',
    long_description='',
    author='Andrei',
    author_email='andrei.zbikowski@gmail.com',
    url='http://github.com/b1naryth1ef/holster',
    packages=packages,
    package_data={},
    package_dir={'holster': 'holster'},
    include_package_data=True,
    install_requires=requires,
    license='Apache 2.0',
    zip_safe=False,
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
    ),
)
